from fileinput import filename
from typing import Optional
import boto3
from botocore.exceptions import NoCredentialsError, ClientError
import json
import logging
import shutil
import os
from mypy_boto3_s3 import S3Client


class S3Helper:
    def __init__(self, folder_path, tmp_path, s3_client, bucket_name):
        self.client: S3Client = s3_client
        self.folder_path = folder_path
        self.bucket_name = bucket_name
        self.keys = self.list_object_keys()
        self.EXAM_DETAILS = None
        self.path = tmp_path
        logging.basicConfig(level=logging.INFO)

    def __enter__(self):
        return self

    # def __exit__(self, exc_type, exc_val, exc_tb):
    #     self.cleanup()

    # def cleanup(self):
    #     if os.path.exists(self.temp_dir):
    #         shutil.rmtree(self.temp_dir)
    #         logging.info(f"Deleted temp directory: {self.temp_dir}")

    def list_object_keys(self):
        """This function lists all the keys at the specified folder path in the S3 bucket.
        function runs when the class is initialized.

        Returns:
            a dictionary with a nested 'Contents' key that hass a list of dictionaries with 'Key' key.
        """
        try:
            response = self.client.list_objects_v2(
                Bucket=self.bucket_name, Prefix=self.folder_path
            )
            keys = [
                obj.get("Key") for obj in response.get("Contents", []) if obj.get("Key")
            ]
            return keys
        except NoCredentialsError:
            logging.error("Credentials not available")
            return []
        except ClientError as e:
            logging.error(f"An error occurred: {e}")
            return []

    def download_file(self, keyword: str) -> str:
        try:
            for key in self.keys:
                if isinstance(key, str) and keyword in key:
                    filename = str(key).split("/")[-1]
                    file_type = filename.split(".")[-1]
                    saved_path = f"{self.path}/{keyword}.{file_type}"
                    self.client.download_file(
                        self.bucket_name, Key=str(key), Filename=saved_path
                    )

                    print(len(saved_path) == 0, "json check length")
                    logging.info(f"Downloaded '{key}' to '{saved_path}'")
                    return saved_path
            raise FileNotFoundError(f"File with keyword {keyword} not found")
        except NoCredentialsError:
            logging.error("Credentials not available")
            raise FileNotFoundError("Credentials not available")
        except ClientError as e:
            logging.error(f"An error occurred: {e}")
            raise FileNotFoundError(f"An error occurred: {e}")

    def download_all_files(self):
        """Download all files from the given list of S3 objects."""
        try:
            for key in self.keys:
                if key:
                    logging.info(f"Downloading file with key: {key}")
                    # filename = f"{self.path}/{key.split('/')[-1]}"
                    filename = f"test/{key.split('/')[-1]}"
                    self.client.download_file(
                        self.bucket_name, Key=str(key), Filename=filename
                    )
        except NoCredentialsError:
            logging.error("Credentials not available")
        except ClientError as e:
            logging.error(f"An error occurred: {e}")
            return False

    def download_exam_ambient_noise(self):
        """Download EXAM_AMBIENT.wav file from S3"""
        return self.download_file("EXAM_AMBIENT")

    def download_ble_stream_motion(self):
        """Download BLE_STREAM_MOTION.wav file from S3"""
        return self.download_file("BLE_STREAM_MOTION")

    def download_ble_stream_audio(self):
        """Download BLE_STREAM_AUDIO.wav file from S3"""
        return self.download_file("BLE_STREAM_AUDIO")

    def download_ble_stream_temperature(self):
        """Download BLE_STREAM_TEMPERATURE.wav file from S3"""
        return self.download_file("BLE_STREAM_TEMPERATURE")

    def download_ble_stream_voltage(self):
        """Download BLE_STREAM_VOLTAGE.wav file from S3"""
        return self.download_file("BLE_STREAM_VOLTAGE")

    def download_exam_video(self):
        """Download EXAM_VIDEO.mp4 file from S3"""
        return self.download_file("EXAM_VIDEO")

    def download_patient_tracking(self):
        """Download PATIENT_TRACKING.json file from S3"""
        return self.download_file("PATIENT_TRACKING")

    def download_exam_details(self):
        """Download EXAM_DETAILS.json file from S3"""
        return self.download_file("EXAM_DETAILS")

    def download_normalized_audio(self):
        """Download NORMALIZED.wav file from S3"""
        return self.download_file("NORMALIZED")

    def load_exam_details(self):
        """Load EXAM_DETAILS.json from S3 to local and parse it"""
        if not self.download_exam_details():
            return False
        try:
            with open("EXAM_DETAILS.json") as f:
                self.EXAM_DETAILS = json.load(f)
            return True
        except FileNotFoundError:
            logging.error("EXAM_DETAILS.json file not found")
        except json.JSONDecodeError:
            logging.error("Error decoding JSON from EXAM_DETAILS.json")
        return False

    def get_exam_details(self):
        """Get keys from the EXAM_DETAILS.json"""
        try:
            if self.EXAM_DETAILS is None:
                new_path = self.download_exam_details()
                if not new_path:
                    return None
                return new_path
            else:

                return self.EXAM_DETAILS
        except FileNotFoundError:
            logging.error("EXAM_DETAILS.json file not found")
            return None

    def get_key_value(self, key):
        """Get value for a specific key from EXAM_DETAILS.json"""
        if self.EXAM_DETAILS is None:
            if not self.load_exam_details():
                return None
        return self.EXAM_DETAILS



from typing_extensions import Buffer
import numpy as np
from scipy.io import wavfile

from typing import Tuple


def load_wav_file(file_path: str) -> Tuple[int, np.ndarray]:
    """
    Load a .wav file and return the sampling rate and data as a NumPy array.

    Args:
    - file_path (str): Path to the .wav file.

    Returns:
    - Tuple[int, np.ndarray]: Sampling rate and the audio data as a NumPy array.
    """
    try:
        sampling_rate, data = wavfile.read(file_path)
        return sampling_rate, data
    except Exception as e:
        raise ValueError(f"An error occurred while loading the .wav file: {e}")


# def load_wav_file_return_audio_buffer_and_sampling_rate(file_path: str) -> Tuple[bytes, int]:
#     """
#     Load a .wav file and return the audio data as a buffer and the sampling rate.

#     Args:
#     - file_path (str): Path to the .wav file.

#     Returns:
#     - Tuple[bytes, int]: Audio data as a buffer and the sampling rate.
#     """
#     try:
#         sampling_rate, data = wavfile.read(file_path)
#         audio_buffer = data.tobytes()
#         return audio_buffer, sampling_rate
#     except Exception as e:
#         raise ValueError(f"An error occurred while loading the .wav file: {e}")



from .s3_helper import S3Helper
from .main import ExamData
from .load_helper import DataInitializer
from .transform_data import load_wav_file


__all__ = ["ExamData", "S3Helper", "DataInitializer", "load_wav_file"]



import boto3
from botocore.exceptions import ClientError
from mypy_boto3_dynamodb import DynamoDBClient
from mypy_boto3_s3.client import S3Client
from typing import Any, Dict, Literal, Tuple
import numpy as np
import json
import logging
import os
import tempfile
import shutil
from typing import TypedDict

from exam_data_helper.aws_setup import dynamodb_client, s3_client
from exam_data_helper.dynamo_helper import DynamoHelper
from exam_data_helper.s3_helper import S3Helper
from exam_data_helper.transform_data import load_wav_file
# from transform_data import load_wav_file
# from s3_helper import S3Helper
# from dynamo_helper import DynamoHelper

# Configure logging
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    handlers=[logging.FileHandler("exam_data.log"), logging.StreamHandler()],
)
logger = logging.getLogger(__name__)

# source types for type hinting
SourceType = Literal["exam_id", "s3", "local"]


class ExamData:
    """No matter what the source type is, the EXAM_DETAILS is always loaded from the S3 bucket.
    - EXAM_DETAILS.json is the source of truth for the exam data
    - if it is an exam_id, the exam_id is used to query the dynamodb table to get the patient_id
    - if it is an s3 path, the path is used to get the EXAM_DETAILS
    - if it is a local directory, the EXAM_DETAILS is loaded from the local directory
    """

    def __init__(
        self, source_type: SourceType, source_value: str, region_name: str = "us-east-2"
    ):
        self.temp_dir = tempfile.mkdtemp()
        logger.info(f"Temporary directory created at {self.temp_dir}")
        self.source_type = source_type
        self.source_value = source_value
        self.region_name = region_name
        self.data: Dict[str, Any] = {}
        self.dynamo_client: DynamoDBClient = dynamodb_client.get_client(region_name)
        self.s3_client: S3Client = s3_client.get_client(region_name)
        self.dynamo_helper = None
        self.s3_helper = S3Helper(
            bucket_name="digibeat-staging-uploads",
            s3_client=self.s3_client,
            folder_path="",
            tmp_path=self.temp_dir,
        )
        self.exam_details_json = None
        

        if source_type == "exam_id":
            self._initialize_from_exam_id(source_value)
        elif source_type == "s3":
            self._initialize_from_s3(source_value)
        elif source_type == "local":
            self._initialize_from_local(source_value)
        else:
            raise ValueError("Invalid source type provided")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup()

    def cleanup(self):
        if os.path.exists(self.temp_dir):
            shutil.rmtree(self.temp_dir)
            logger.info(f"Temporary directory {self.temp_dir} has been removed")

    def _initialize_from_exam_id(self, exam_id: str):
        try:
            logger.info(f"Data initialized using exam ID: {exam_id}")
            self.dynamo_helper = DynamoHelper(self.dynamo_client)
            response = self.dynamo_helper.get_exam_details(exam_id)
            self.data.update(response)
            # self.data[response] = response
            s3_path = f"PATIENT#{self.data['PATIENT_ID']}/EXAM#{exam_id}"
            self._initialize_from_s3(s3_path)
        except ClientError as e:
            logger.error(f"An error occurred: {e}")
            raise

    def _initialize_from_s3(self, s3_path):
        exam_details_path = self.s3_helper.get_exam_details()

        if exam_details_path != None:
            # self.exam_details_json = exam_details_path
            with open(exam_details_path, "r") as f:
                self.exam_details_json = json.load(f)

        logger.info(f"Data loaded from S3: {s3_path}")

    def _initialize_from_local(self, local_dir: str):
        try:
            for root, _, files in os.walk(local_dir):
                for file in files:
                    file_path = os.path.join(root, file)
                    if file.endswith(".json"):
                        with open(file_path, "r") as f:
                            self.data.update(json.load(f))
                    else:
                        shutil.copy(file_path, self.temp_dir)
            logger.info(f"Data loaded from local directory: {local_dir}")
        except Exception as e:
            logger.error(f"Error loading data from local directory: {e}")
            raise

    def download_exam_ambient_noise(self) -> str:
        try:
            if (self.source_type == "exam_id") or (self.source_type == "s3"):
                result = self.s3_helper.download_exam_ambient_noise()
                if result != None:
                    return result
                else:
                    raise FileNotFoundError("Ambient noise file not found")
            elif self.source_type == "local":
                logger.info("Ambient noise download not supported for local data")
            raise ValueError("Invalid source type provided")
        except Exception as e:
            logger.error(f"An error occurred while downloading ambient noise: {e}")
            raise

    def get_exam_voltage_data(self) -> Tuple[int, np.ndarray]:
        """returns the voltage data from the exam in the form of a numpy array and the sampling rate

        Raises:
            FileNotFoundError: file not found
            ValueError: invalid source type

        Returns:
            Tuple[int, np.ndarray]: numpy array and sampling rate
        """
        try:
            if (self.source_type == "exam_id") or (self.source_type == "s3"):
                result_path = self.s3_helper.download_ble_stream_voltage()
                if result_path != None:
                    return load_wav_file(result_path)
                else:
                    raise FileNotFoundError("Voltage data file not found")
            elif self.source_type == "local":
                logger.info("Voltage data download not supported for local data")
            raise ValueError("Invalid source type provided")
        except Exception as e:
            logger.error(f"An error occurred while downloading voltage data: {e}")
            raise

    def get_exam_audio_data(self) -> Tuple[int, np.ndarray]:
        """returns the audio data from the exam in the form of a buffer and the sampling rate

        Raises:
            FileNotFoundError: file not found
            ValueError: invalid source type

        Returns:
            Tuple[Buffer, int]: buffer and sampling rate
        """
        try:
            if (self.source_type == "exam_id") or (self.source_type == "s3"):
                result_path = self.s3_helper.download_ble_stream_audio()
                if result_path != None:
                    return load_wav_file(result_path)
                else:
                    raise FileNotFoundError("Audio data file not found")
            elif self.source_type == "local":
                logger.info("Audio data download not supported for local data")
            raise ValueError("Invalid source type provided")
        except Exception as e:
            logger.error(f"An error occurred while downloading audio data: {e}")
            raise

    def get_exam_details(self) -> Dict[str, Any] | None:
        """returns the exam details

        Returns:
            Dict[str, Any] | None: exam details
        """
        return self.exam_details_json



import os
import logging
from botocore.exceptions import ClientError
from exam_data_helper.aws_setup import s3_client

logger = logging.getLogger(__name__)


class DataInitializer:
    def __init__(
        self, region_name: str = "us-east-2"
    ):
        self.region_name = region_name

    @staticmethod
    def from_local_directory(directory: str) -> dict:
        """Initialize data from a local directory."""
        data = {}
        try:
            for filename in os.listdir(directory):
                filepath = os.path.join(directory, filename)
                if os.path.isfile(filepath):
                    with open(filepath, "r") as file:
                        data[filename] = file.read()
            logger.info(f"Data initialized from local directory: {directory}")
        except Exception as e:
            logger.error(f"An error occurred while reading from local directory: {e}")
            raise
        return data
    
    def from_s3_directory(self, bucket: str, directory: str) -> dict:
        """Initialize data from an S3 directory."""
        data = {}
        try:
            client = s3_client.get_client(self.region_name)
            response = client.list_objects_v2(Bucket=bucket, Prefix=directory)
            for obj in response.get("Contents", []):
                key = obj.get("Key")
                if key and not key.endswith("/"):
                    file_obj = client.get_object(Bucket=bucket, Key=key)
                    data[key] = file_obj["Body"].read().decode("utf-8")
            logger.info(
                f"Data initialized from S3 bucket: {bucket}, directory: {directory}"
            )
        except ClientError as e:
            logger.error(f"An error occurred while reading from S3: {e}")
            raise
        return data

    @staticmethod
    def from_stream(multipart_upload_stream_obj) -> dict:
        """Initialize data from a stream (e.g., multipart upload)."""
        data = {}
        try:
            data["stream_data"] = multipart_upload_stream_obj.read().decode("utf-8")
            logger.info("Data initialized from stream.")
        except Exception as e:
            logger.error(f"An error occurred while reading from stream: {e}")
            raise
        return data



# dynamodb_client.py
import boto3
from mypy_boto3_dynamodb import DynamoDBClient
from mypy_boto3_s3.client import S3Client

class dynamodb_client:
    _client = None

    @classmethod
    def get_client(cls, region_name: str) -> DynamoDBClient:
        if cls._client is None:
            cls._client = boto3.client("dynamodb", region_name=region_name)
        return cls._client
    
class s3_client:
    _client = None

    @classmethod
    def get_client(cls, region_name: str) -> S3Client:
        if cls._client is None:
            cls._client = boto3.client("s3", region_name=region_name)
        return cls._client



import boto3
from botocore.exceptions import NoCredentialsError, ClientError
from mypy_boto3_dynamodb import DynamoDBClient
from boto3.dynamodb.types import TypeDeserializer, TypeSerializer
from typing import Any, Dict
import logging


class DynamoHelper:
    """This class helps to interact with the DynamoDB table that contains the exam data."""

    def __init__(self, dynamo_client):
        self.client: DynamoDBClient = dynamo_client
        # self.table_name = table_name
        self.deserializer = TypeDeserializer()
        self.serializer = TypeSerializer()
        logging.basicConfig(level=logging.INFO)

    def get_exam_details(self, exam_id: str) -> Dict[str, Any]:
        """This function gets the exam details from the DynamoDB table.

        Args:
            exam_id: The exam_id for which the details are to be fetched.

        Returns:
            A dictionary with the exam details.
        """
        try:
            data = {}
            response = self.client.query(
                TableName="main-staging",
                ExpressionAttributeValues={
                    ":PK": {"S": f"EXAM#{exam_id}"},
                    ":SK": {"S": "PROFILE"},
                },
                KeyConditionExpression="PK = :PK and SK = :SK",
            )
            deserializer = TypeDeserializer()
            for item in response["Items"]:
                for key, value in item.items():
                    data[key] = deserializer.deserialize(value)
            return data
        except NoCredentialsError:
            logging.error("Credentials not available")
            return {}
        except ClientError as e:
            logging.error(f"An error occurred: {e}")
            return {}



from exam_data_helper import ExamData


def test_s3_helper():
    ######## TESTING s3 filepath method here
    # exam_data = ExamData(source_type='s3', source_value='PATIENT#1165/EXAM#1188')
    # res = exam_data.get_exam_voltage_data()
    # print(res, 'res')

    # # # s3.download_exam_video()
    # res = s3.get_json_keys()
    # print(res, 'res')

    ######## TESTING exam_id method here
    exam_data = ExamData(source_type="exam_id", source_value="1193", region_name="us-east-2")
    res = exam_data.get_exam_voltage_data()
    json = exam_data.get_exam_details()
    print(res, "res")
    print(json, "json")
    # exam_data = ExamData(source_type='exam_id', source_value='1188')
    # test = exam_data.download_exam_ambient_noise()
    # print(test, 'test')

    ######## TESTING local method here


if __name__ == "__main__":
    test_s3_helper()
